"use strict";
var opt = require('node-getopt').create([
    ['m', 'operator=OP+', 'Morphological operator(s) (default: dilationSquare-20).'],
    ['f', 'frame=FR+', 'Frame(s) to use (default: basisRGB).'],
    ['', 'nogamma', 'Do not use gamma correction (default: correct for sRGB).'],
    ['c', 'clip', 'Clip channels independently (default: preserve hue).'],
    ['o', 'output=OUT', 'Output file (only valid with a single output).'],
    ['s', 'stem=STEM', 'Stem of outputs (default: input file without extension).'],
    ['v', 'verbose', 'Be verbose in generating output file names (default: no).'],
    ['h' , 'help'                , 'display this help'],
]).bindHelp(
    "Usage: node filter.js [OPTIONS] INPUT_IMAGE1.png\n" +
    "  Morphological filters using group-invariant frames.\n" +
    "\n" +
    "[[OPTIONS]]\n" +
    "\n" +
    "The following operators are available out-of-the-box:\n" +
    "  dilationSquare, erosionSquare, openSquare, closeSquare, occoSquare\n" +
    "For the -m/--operator option, append a dash and the structuring element size. " +
    "Note that a structuring element of size 2 will cover three pixels (the 2 refers to the distance between pixel centers).\n" +
    "\n" +
    "The following frames are available out-of-the-box:\n" +
    "  basisRGB, satInvFrameRGB, hueInvFrameRGB, sathueInvFrameRGB, rotInvFrameRGB, satInvFrameSVG, hueInvFrameSVG, sathueInvFrameSVG\n"
).parseSystem();
var fs = require('fs');
var png = require('pngjs').PNG;
var frameMorphology = require('frameMorphology');

// Validate and process options
var inputRegExp = /^(.*)\.png$/i;

if (opt.argv.length!=1) { // Check this before determining the stem (at least), as it uses argv.
    console.error("One, and only one, input file is required.\n");
    process.exit(1);
}

if (!inputRegExp.test(opt.argv[0])) {
    console.error("Input filename must end in .png (and the file must in fact be a PNG file).\n");
    process.exit(2);
}

var operators = (opt.options.operator && opt.options.operator.length>0) ? opt.options.operator : ['dilationSquare-20'];
var frames = (opt.options.frame && opt.options.frame.length>0) ? opt.options.frame : ['basisRGB'];
var gammacorrect = !opt.options.nogamma;
var clipping = opt.options.clip ? 'independentClipping' : 'centroidClipping';
var output = opt.options.output;
var stem = opt.options.stem ? opt.options.stem : inputRegExp.exec(opt.argv[0])[1];
var verbose = opt.options.verbose;

if (output && operators.length*frames.length>1) { // Do this after finding the operators and frames, to take the defaults into account.
    console.error("The -o/--output option can only be used when ONE operator and ONE frame are given.\n");
    process.exit(3);
}

// Actual filtering
var op, fr;

function constructFilterCallback(operation, frame) {
    return function() {
        var outputName = output ? output : verbose ? ([stem, operation, frame, gammacorrect ? 'linearRGB' : 'sRGB', clipping].join('-') + '.png') : ([stem, operation, frame].join('-') + '.png');
        frameMorphology.filter(this, operation, frame, gammacorrect, clipping);
        this.pack().pipe(fs.createWriteStream(outputName));
    };
}

for(op=0; op<operators.length; op++) {
    for(fr=0; fr<frames.length; fr++) {
        fs.createReadStream(opt.argv[0])
            .pipe(new png())
            .on('parsed', constructFilterCallback(operators[op], frames[fr]));
    }
}

