"use strict";

function lessThanOrEqualTo(a,b) { return a <= b; }
function greaterThanOrEqualTo(a,b) { return a >= b; }

// These methods are based on:
//   Computationally efficient, one-pass algorithm for morphological filters Journal of Visual Communication and Image Representation, Vol. 22, No. 5. (July 2011), pp. 411-420, doi:10.1016/j.jvcir.2011.03.005 by Petr Dokládal, Eva Dokládalová

// This is the "class" used to implement the required queue, the actual operators follow below.
function MorphologicalQueue(size,cmp) {
    var l2 = 0, s = 1;
    while(l2<31 && s<size) { // Compute smallest power-of-two greater than or equal to "size".
        l2++;
        s = s<<1;
    }
    this.size = size>>0; // Truncate size, essentially a hint to the compiler that size is supposed to be an integer.
    this.queue = new Array(2<<l2); // Two elements per item (a position and a value). The array is used circularly, as there are never more than "size" items in the queue. We use a power-of-two to make the modulo operation easier.
    this.queueBegin = 0; // These indices are /not/ used circularly, as this makes the logic easier. So instead of keeping these indices within the bounds of the array, every access to the array is guarded.
    this.queueEnd = 0;
    this.cmp = cmp;
}
MorphologicalQueue.prototype.firstVal = function () {
    return this.queueBegin<this.queueEnd ? this.queue[(this.queueBegin+1)&(this.queue.length-1)] : undefined;
};
MorphologicalQueue.prototype.empty = function () {
    this.queueBegin = 0;
    this.queueEnd = 0;
};
MorphologicalQueue.prototype.enqueue = function (pos, val) {
    var q = this.queue, cmp = this.cmp, lengthMask = q.length-1;

    // Unshift first two elements if they are falling off the window.
    if (this.queueBegin<this.queueEnd && pos >= q[this.queueBegin&lengthMask]) {
        this.queueBegin += 2;
    }

    if (val!==undefined) {
        // Pop elements as long as they are "useless".
        while(this.queueBegin<this.queueEnd && cmp(val,q[(this.queueEnd+lengthMask)&lengthMask])) { // (end+lengthMask)&lengthMask = (end-1) modulo q.length
            this.queueEnd -= 2;
        }
        
        // Push new element.
        q[(this.queueEnd++)&lengthMask] = pos+this.size;
        q[(this.queueEnd++)&lengthMask] = val;
    }
};

// The following operators are available:
//   erosion-, dilation-, open-, close-, occoSquare: erosion (etc.) by a square
// The operators support the following arguments:
//   data: array with data (row-major, each "pixel" only has a single (grey) value)
//   width, height: size of image
//   length: total length of one side of the square
// The length parameter is rounded to the nearest odd integer for erosionSquare and dilationSquare, such that even integers are rounded up.
// For the opening, closing and occo operators the length parameter is rounded to the smallest integer strictly larger than the given length to determine the actual pixel size of the SE.
// The motivation for these choices is that a SE with a length of 2 pixels touches three pixel centers.

exports.erosionSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var radius = Math.floor(lengthFrac/2);
    var xqueue = new MorphologicalQueue(2*radius+1, lessThanOrEqualTo);
    var yqueues = new Array(width+radius);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(2*radius+1, lessThanOrEqualTo);
    }
    for(y=0; y<height+radius; y++) {
        xqueue.empty();
        for(x=0; x<width+radius; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            if (y>=radius && x>=radius) {
                data[(y-radius)*width+(x-radius)] = yqueues[x].firstVal();
            }
        }
    }
    return data;
}

exports.dilationSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var radius = Math.floor(lengthFrac/2);
    var xqueue = new MorphologicalQueue(2*radius+1, greaterThanOrEqualTo);
    var yqueues = new Array(width+radius);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(2*radius+1, greaterThanOrEqualTo);
    }
    for(y=0; y<height+radius; y++) {
        xqueue.empty();
        for(x=0; x<width+radius; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            if (y>=radius && x>=radius) {
                data[(y-radius)*width+(x-radius)] = yqueues[x].firstVal();
            }
        }
    }
    return data;
}

exports.closeSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var length = Math.floor(lengthFrac);
    var xqueue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var xqueue2 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var yqueues = new Array(width+length);
    var yqueues2 = new Array(width+length);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
        yqueues2[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    }
    for(y=0; y<height+length; y++) {
        xqueue.empty();
        xqueue2.empty();
        for(x=0; x<width+length; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            xqueue2.enqueue(x, yqueues[x].firstVal());
            yqueues2[x].enqueue(y, xqueue2.firstVal());
            if (y>=length && x>=length) {
                data[(y-length)*width+(x-length)] = yqueues2[x].firstVal();
            }
        }
    }
    return data;
}

exports.openSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var length = Math.floor(lengthFrac);
    var xqueue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var xqueue2 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var yqueues = new Array(width+length);
    var yqueues2 = new Array(width+length);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
        yqueues2[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    }
    for(y=0; y<height+length; y++) {
        xqueue.empty();
        xqueue2.empty();
        for(x=0; x<width+length; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
            }
            yqueues[x].enqueue(y, xqueue.firstVal());
            xqueue2.enqueue(x, yqueues[x].firstVal());
            yqueues2[x].enqueue(y, xqueue2.firstVal());
            if (y>=length && x>=length) {
                data[(y-length)*width+(x-length)] = yqueues2[x].firstVal();
            }
        }
    }
    return data;
}

exports.occoSquare = function (data, width, height, lengthFrac) {
    var x, y;
    var length = Math.floor(lengthFrac);
    var xqueue = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var xqueue2 = new MorphologicalQueue(2*length+1, lessThanOrEqualTo); // Combine erosion of the closing with erosion of the opening
    var xqueue3 = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
    var xdqueue = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var xdqueue2 = new MorphologicalQueue(2*length+1, greaterThanOrEqualTo); // Combine dilation of the erosion with dilation of the closing
    var xdqueue3 = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    var yqueues = new Array(width+2*length);
    var yqueues2 = new Array(width+2*length);
    var yqueues3 = new Array(width+2*length);
    var ydqueues = new Array(width+2*length);
    var ydqueues2 = new Array(width+2*length);
    var ydqueues3 = new Array(width+2*length);
    for(x=0; x<yqueues.length; x++) {
        yqueues[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
        yqueues2[x] = new MorphologicalQueue(2*length+1, lessThanOrEqualTo);
        yqueues3[x] = new MorphologicalQueue(length+1, greaterThanOrEqualTo);
        ydqueues[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
        ydqueues2[x] = new MorphologicalQueue(2*length+1, greaterThanOrEqualTo);
        ydqueues3[x] = new MorphologicalQueue(length+1, lessThanOrEqualTo);
    }
    for(y=0; y<height+2*length; y++) {
        xqueue.empty();
        xqueue2.empty();
        xqueue3.empty();
        xdqueue.empty();
        xdqueue2.empty();
        xdqueue3.empty();
        for(x=0; x<width+2*length; x++) {
            if (y<height && x<width) {
                xqueue.enqueue(x, data[y*width+x]);
                xdqueue.enqueue(x, data[y*width+x]);
            } else {
                xqueue.enqueue(x);
                xdqueue.enqueue(x);
            }
            
            yqueues[x].enqueue(y, xqueue.firstVal());
            xqueue2.enqueue(x, yqueues[x].firstVal());
            yqueues2[x].enqueue(y, xqueue2.firstVal());
            xqueue3.enqueue(x, yqueues2[x].firstVal());
            yqueues3[x].enqueue(y, xqueue3.firstVal());
            
            ydqueues[x].enqueue(y, xdqueue.firstVal());
            xdqueue2.enqueue(x, ydqueues[x].firstVal());
            ydqueues2[x].enqueue(y, xdqueue2.firstVal());
            xdqueue3.enqueue(x, ydqueues2[x].firstVal());
            ydqueues3[x].enqueue(y, xdqueue3.firstVal());
            
            if (y>=2*length && x>=2*length) {
                data[(y-2*length)*width+(x-2*length)] = 0.5*(yqueues3[x].firstVal() + ydqueues3[x].firstVal());
            }
        }
    }
    return data;
}

